#ifndef _LINUX_TIVO_PRIVATE_H
#define _LINUX_TIVO_PRIVATE_H

#define IDETIVO_PROFILE 

#define SECTOR_BYTES 512
#define MAX_SECTORS 256 

#include <linux/tivo-scramble.h>
#include <linux/tivo-monotime.h>

struct idetivo_driver_info; 
struct idetivo_driver_info *idetivo_init_driver_info(void);

/*
 * Special request format for TiVo volumes
 */

enum idetivo_request_status { IDETIVO_FREE, IDETIVO_DONE, IDETIVO_BUSY,
			      IDETIVO_ERROR }; 

enum FsIovecMode { IOVEC_READ, IOVEC_WRITE }; 

struct FsIovecState {
    
    enum FsIovecMode mode; /* read or write */

    /* This is for the DMA engine, what it thinks the top is */
    struct FsIovec *top; 
    
    /* This is the original Iovec, we promise not to touch it... */
    struct FsIovec *head;
    int head_len; 

    /* This is for the FsIovecTransfer code... where it will start writing
       out its bytes */
    struct FsIovec *curr_iovec; /* the currently "active" iovec */
    int curr_iovec_bytes; /* number of bytes into the "active" iovec */
    int total_bytes; /* total number of bytes so far */
};

struct idetivo_request {
    int cmd; /* READ or WRITE */
    struct ide_drive_s *drive;
    enum idetivo_request_status status; /* { free, done, busy } */
    unsigned long int sector; /* starting sector number */
    unsigned int num_sectors; /* total number of sectors to transfer */
    unsigned int chunk_num_sectors; /* number of sectors in this chunk */
    unsigned int partition_flags; /* flags from partition being transferred */

    struct FsIovecState state; 

    int drive_status;   /* Drive status upon completion */
    int progress; /* total progress in bytes */
    int sector_progress; /* total number of sectors transferred */
    int chunk_sector_progress; /* progress inside this sector chunk */
    
    monotonic_t deadline;
    struct idetivo_request *next; /* next in the elevator-ordered list */
    struct idetivo_request *next_rt; /* next in the deadline ordered list */
    struct task_struct *task; /* the task that made this request */
    struct wait_queue *wait; /* wait for this request to finish */
    struct io_scramble scramble; /* scrambler keying */

#ifdef IDETIVO_PROFILE
    unsigned long iostart; /* time in jiffies of I/O start */
    int num_interrupts; 
#endif 
};

#define DISK_PERFORMANCE_SIZE 64
#define DISK_PERF_LOOKUP(perf, sector) (&((perf)[((sector) >> 18)]))
#define REQUEST_POOL_SIZE 32

struct disk_performance {
    int seek;
    int transfer; 
};

enum media_state {
   STATE_UNKNOWN, STATE_KNOWN, STATE_MEDIA, STATE_NOMEDIA, STATE_FLUSHED
};

#define IDETIVO_DRIVER_VERSION 0x00010004

enum drive_family {
   FAMILY_UNKNOWN, FAMILY_QUANTUM, FAMILY_STREAMWEAVER1, FAMILY_UNSUPPORTED
};

struct idetivo_driver_info {

    struct disk_performance perf[DISK_PERFORMANCE_SIZE]; 
    
    struct idetivo_request request_pool[REQUEST_POOL_SIZE];
    struct idetivo_request *request_head; /* head of all elevator-ordered
                                             lists */
    struct idetivo_request *request_head_rt; /* head of all deadline ordered
                                                lists */
    struct idetivo_request *cur; /* request currently being serviced */
    int num_requests; /* number of requests in the lists */
    struct wait_queue *wait_for_request;

#ifdef IDETIVO_PROFILE
    /* This is just for performance monitoring... */
    int total_requests;
    unsigned long long total_jiffies;
    unsigned long long total_bytes;
    int elev;
    int elev_rt;
    int rt_only; 
#endif

   /* These two fields are legacy - no longer used, but they can't
      be removed with complete safety.  Leave 'em in. */
   
   enum media_state state;
   char config[512];

    /* Here's a page of memory that you can use for temporary storage...
       (we have to copy from user memory to kernel memory before sending
       data to the drive, and thats one thing this is used for) */
#define TEMP_PAGE_SIZE 512
    char temp_page[TEMP_PAGE_SIZE];

    union {
        struct { long block; } lba; 
        struct {
            long cylinder;
            long head;
            long sector;
        } chs; 
    } head_pos;
   
   /*
    * Stick the version code here, after everything which was created
    * prior to 2/17/2000, so it doesn't interfere with older versions
    * of ideturbo.
    */

    unsigned long int version;
   
};

/*
 * Create per-drive private-state information, in addition to the
 * per-hardware-interface information in idetivo_driver_info.
 */
  
struct idetivo_drive_info {
   enum drive_family family;

   /* 
    * State-switching must be done on a per-drive basis - state and
    * config array in idetivo_driver_info fields are no longer used.
    */
   
   enum media_state state;
   char config[512];
 
   int back_to_back_writes;
   int back_to_back_limit;

};


/* ////////////////////////////////////////////////////////////////////
// 
// function prototypes
// 
//////////////////////////////////////////////////////////////////// */

void idetivo_release_request(ide_drive_t *drive, 
                             struct idetivo_request *which);


void idetivo_cancel_request(ide_drive_t *drive,
                            struct idetivo_request *which);

struct FsDiskState {
    int c; /* cylinder */
    int h; /* head */
    int s; /* sector */
}; 

void FsIovecInit(struct FsIovecState *state,
                 struct FsIovec *head,
                 int head_len,
                 enum FsIovecMode mode);
int FsIovecTransfer(struct FsIovecState *state,
                    char *ptr, int bytes);

/*
 * Ok, this is an expanded form so that we can use the same
 * request for paging requests when that is implemented. In
 * paging, 'bh' is NULL, and the semaphore is used to wait
 * for read/write completion.
 */
struct generic_request {
    struct hd_generic_cmd cmd;
    struct FsIovecState state;
/*    struct wait_queue *wait; */
/*    struct task_struct *task; */
/*    int status;  */
};

#endif
