/*
 *  linux/mm/vmscan.c
 *
 *  Copyright (C) 1991, 1992, 1993, 1994  Linus Torvalds
 *
 *  Swap reorganised 29.12.95, Stephen Tweedie.
 *  kswapd added: 7.1.96  sct
 *  Version: $Id: //depot/mainline/linux-2.1/mm/vmscan.c#12 $
 */

#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/head.h>
#include <linux/kernel.h>
#include <linux/kernel_stat.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/swap.h>
#include <linux/fs.h>
#include <linux/swapctl.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>

#include <asm/dma.h>
#include <asm/system.h> /* for cli()/sti() */
#include <asm/uaccess.h> /* for copy_to/from_user */
#include <asm/bitops.h>
#include <asm/pgtable.h>
#ifdef CONFIG_XMON
#include <asm/process.h>    // For process logging stuff...
#endif

#undef FLUSH_DEBUG

extern void fnoaccess(unsigned long, int, int, struct task_struct *);

/* 
 * When are we next due for a page scan? 
 */
static int next_swap_jiffies = 0;

/* 
 * How often do we do a pageout scan during normal conditions?
 * Default is four times a second.
 */
int swapout_interval = HZ / 4;

/* 
 * The wait queue for waking up the pageout daemon:
 */
static struct wait_queue * kswapd_wait = NULL;

/* 
 * We avoid doing a reschedule if the pageout daemon is already awake;
 */
static int kswapd_awake = 0;

/*
 * sysctl-modifiable parameters to control the aggressiveness of the
 * page-searching within the kswapd page recovery daemon.
 */
kswapd_control_t kswapd_ctl = {4, -1, -1, -1, -1};

/*
  TiVo special stats
*/
swap_stat_tivo_t swap_stat_tivo;

static void init_swap_timer(void);

/*
 * The swap-out functions return 1 if they successfully
 * threw something out, and we got a free page. It returns
 * zero if it couldn't do anything, and any other value
 * indicates it decreased rss, but the page was shared.
 *
 * NOTE! If it sleeps, it *must* return 1 to make sure we
 * don't continue with the swap-out. Otherwise we may be
 * using a process that no longer actually exists (it might
 * have died while we slept).
 *
 * ALSO! Don't keep any state after sleeping; the process
 * might exit while we were out.
 *
 * The VM_FLUSH flag is set by a process flushing it's
 * own stuff out of memory, so we know he won't go away
 * while it's sleeping, so it's OK to just loop through
 * the whole shebang. Other processes trying to touch
 * such a process will not, since they honor the flag.
 */
static inline int try_to_swap_out(struct task_struct * tsk, struct vm_area_struct* vma,
	unsigned long address, pte_t * page_table, int dma, int wait)
{
	pte_t pte;
	unsigned long entry;
	unsigned long page;
	struct page * page_map;

	pte = *page_table;
	if (!pte_present(pte))
		return 0;
	page = pte_page(pte);
	if (MAP_NR(page) >= max_mapnr) {
#ifdef FLUSH_DEBUG
            if (vma->vm_flags & VM_FLUSH)
                printk("Crazy page at 0x%lx: %ld\n",
                    address, MAP_NR(page), pte);
#endif
            return 0;
        }

#ifdef FLUSH_DEBUG
	if (vma->vm_flags & VM_FLUSH)
		printk("Flush page: 0x%lx page %4ld pte 0x%p ",
			address, MAP_NR(page), pte);
#endif

	page_map = mem_map + MAP_NR(page);
	if (PageReserved(page_map)
	    || PageLocked(page_map)
	    || (dma && !PageDMA(page_map))) {
#ifdef FLUSH_DEBUG
                if (vma->vm_flags & VM_FLUSH)
                        printk("Ignoring guarded page\n");
#endif
		return 0;
        }

        if (!(vma->vm_flags & VM_FLUSH)) {
            /* Deal with page aging.  Pages age from being unused; they
             * rejuvenate on being accessed.  Only swap old pages (age==0
             * is oldest). */
            if ((pte_dirty(pte) && delete_from_swap_cache(MAP_NR(page))) 
                || pte_young(pte))  {
                    set_pte(page_table, pte_mkold(pte));
                    touch_page(page_map,tsk);
                    return 0;
            }
            age_page(page_map,tsk);
            if (page_map->age) {
	       if (tsk->policy != SCHED_OTHER &&
		   page_map->age < swap_control.sc_rt_saving_throw) {
		  /*
		   * Flush the TLB here, to increase the odds that the
		   * realtime process which owns this page will take
		   * a TLB miss before the page is actually swapped, and
		   * will thus mark the page as young and won't swap it.
		   */
		  swap_stat_tivo.sst_realtime_saving_throw ++;
		  flush_tlb_page(vma, address);
	       }
	       return 0;
	    }
        }
        if (pte_dirty(pte)) {
            delete_from_swap_cache(MAP_NR(page));
            if (vma->vm_ops && vma->vm_ops->swapout) {
                    pid_t pid = tsk->pid;
                    vma->vm_mm->rss--;
#ifdef CONFIG_XMON
                    _pl_filemap_page_swap(tsk->pid, page, address);
#endif
                    if (vma->vm_ops->swapout(vma, address - vma->vm_start + vma->vm_offset,
                            page_table))
                            kill_proc(pid, SIGBUS, 1);
            } else {
                    if (page_map->count != 1) {
#ifdef CONFIG_XMON
                        _pl_generic_event2(PLOG_TYPE_SHARED_PAGE_IGNORED, 
                                tsk->pid, address);
#endif
                            return 0;
		    }
                    if (!(entry = get_swap_page())) {
#ifdef CONFIG_XMON
                        _pl_generic_event2(PLOG_TYPE_NO_SWAP_PAGE, tsk->pid, 
                                address);
#endif
                            return 0;
		    }
                    vma->vm_mm->rss--;
                    flush_cache_page(vma, address);
                    set_pte(page_table, __pte(entry));
                    flush_tlb_page(vma, address);
                    tsk->nswap++;
#ifdef CONFIG_XMON
                    _pl_generic_page_swap(tsk->pid, page, address);
#endif
                    rw_swap_page(WRITE, entry, (char *) page, wait);
            }
            free_page(page);
            return 1;	/* we slept: the process may not exist any more */
	}
        if ((entry = find_in_swap_cache(MAP_NR(page)))) {
		if (page_map->count != 1) {
#ifdef CONFIG_XMON
                    _pl_generic_event2(PLOG_TYPE_DIRTY_MULTIPLE_NO_SWAP, 
                            tsk->pid, address);
#endif

                    // This condition means that the direct DMA to user space
                    // code has added a reference to the page.  Since
                    // find_in_swap_cache() actually removed the swap cache
                    // entry, we now restore it.  I don't think this should
                    // ever fail (we've done nothing that could cause a context
                    // switch...)
                    if (!add_to_swap_cache(MAP_NR(page), entry)) {
                        debug("Ack!  Failed to restore swap cache entry???", 0);
                    }
                    return 0;
		}
		vma->vm_mm->rss--;
		flush_cache_page(vma, address);
#ifdef CONFIG_XMON
                _pl_reswap(tsk->pid, page, address);
#endif
		set_pte(page_table, __pte(entry));
		flush_tlb_page(vma, address);
		free_page(page);
		return 1;
	} 
	vma->vm_mm->rss--;
	flush_cache_page(vma, address);
	pte_clear(page_table);
	flush_tlb_page(vma, address);
	entry = page_unuse(page);
#ifdef CONFIG_XMON
        _pl_pageout(tsk->pid, page, address);
#endif
	free_page(page);
	return entry;
}

/*
 * A new implementation of swap_out().  We do not swap complete processes,
 * but only a small number of blocks, before we continue with the next
 * process.  The number of blocks actually swapped is determined on the
 * number of page faults, that this process actually had in the last time,
 * so we won't swap heavily used processes all the time ...
 *
 * Note: the priority argument is a hint on much CPU to waste with the
 *       swap block search, not a hint, of how much blocks to swap with
 *       each process.
 *
 * (C) 1993 Kai Petzke, wpp@marie.physik.tu-berlin.de
 */

static inline int swap_out_pmd(struct task_struct * tsk, struct vm_area_struct * vma,
	pmd_t *dir, unsigned long address, unsigned long end, int dma, int wait)
{
	pte_t * pte;
	unsigned long pmd_end;
        extern int guard_if_unguarded(unsigned long, int, struct task_struct *);

	if (pmd_none(*dir))
		return 0;
	if (pmd_bad(*dir) ||
	    ((pmd_val(*dir) & 0x80000000) != 0x80000000)) {
		extern void debug(char *, struct pt_regs *);
		printk("swap_out_pmd: bad pmd (%08lx)\n", pmd_val(*dir));
		printk("task %p vma %p pmd %p addr 0x%08lx end 0x%08lx\n",
			tsk, vma, dir, address, end);
		debug("Bad VM Puppy", 0);
		pmd_clear(dir);
		return 0;
	}
	
	pte = pte_offset(dir, address);
	
	pmd_end = (address + PMD_SIZE) & PMD_MASK;
	if (end > pmd_end)
		end = pmd_end;

	do {
		int result;
                int guarded = 0;
		tsk->swap_address = address + PAGE_SIZE;

                if (vma->vm_inode) {
                    if (!guard_if_unguarded(address, PAGE_SIZE, tsk))
                        guarded = 1;
                    else
                        goto nextpage;
                }

                result = try_to_swap_out(tsk, vma, address, pte, dma, wait);

                if (guarded)
                    fnoaccess(address, PAGE_SIZE, 0, tsk);

		if (!(vma->vm_flags & VM_FLUSH) && result)
			return result;

		if (need_resched) {
			if (!(vma->vm_flags & VM_FLUSH))
				return -1;
			schedule();
                        // We are allowed to continue here because the
                        // VM_FLUSH flag tells us that we are swapping out our
                        // *own* address space, so it must still be valid.
		}

        nextpage:
		address += PAGE_SIZE;
		pte++;
	} while (address < end);
	return 0;
}

static inline int swap_out_pgd(struct task_struct * tsk, struct vm_area_struct * vma,
	pgd_t *dir, unsigned long address, unsigned long end, int dma, int wait)
{
	pmd_t * pmd;
	unsigned long pgd_end;

	if (pgd_none(*dir))
		return 0;
	if (pgd_bad(*dir)) {
		printk("swap_out_pgd: bad pgd (%08lx)\n", pgd_val(*dir));
		pgd_clear(dir);
		return 0;
	}

	pmd = pmd_offset(dir, address);

	pgd_end = (address + PGDIR_SIZE) & PGDIR_MASK;	
	if (end > pgd_end)
		end = pgd_end;
	
	do {
		int result = swap_out_pmd(tsk, vma, pmd, address, end, dma, wait);
		if (!(vma->vm_flags & VM_FLUSH) && result)
			return result;
		address = (address + PMD_SIZE) & PMD_MASK;
		pmd++;
		if (need_resched) {
			if (!(vma->vm_flags & VM_FLUSH))
				return -1;
			schedule();
                        // We are allowed to continue here because the
                        // VM_FLUSH flag tells us that we are swapping out our
                        // *own* address space, so it must still be valid.
		}
	} while (address < end);
	return 0;
}

static int swap_out_vma(struct task_struct * tsk, struct vm_area_struct * vma,
	pgd_t *pgdir, unsigned long start, int dma, int wait)
{
	unsigned long end;

	/* Don't swap out areas like shared memory which have their
	    own separate swapping mechanism or areas which are locked down */
	if (vma->vm_flags & (VM_SHM | VM_LOCKED))
		return 0;

        /*
         * XXX TiVo hack: if a file lock is outstanding on the associated inode,
         *     don't swap the region.
         *
        if (vma->vm_inode && vma->vm_inode->i_flock)
                return 0;
	*/

#ifdef FLUSH_DEBUG
	if (vma->vm_flags & VM_FLUSH)
	    printk("Swapping VMA 0x%p: 0x%lx to 0x%lx (%ld pages)\n",
            vma, start, vma->vm_end,
            (vma->vm_end - start) / PAGE_SIZE);
#endif

	end = vma->vm_end;
	while (start < end) {
		int result = swap_out_pgd(tsk, vma, pgdir, start, end, dma, wait);
		if (!(vma->vm_flags & VM_FLUSH) && result) {
			return result;
                }
		start = (start + PGDIR_SIZE) & PGDIR_MASK;
		pgdir++;
		if (need_resched) {
			if (!(vma->vm_flags & VM_FLUSH))
				return -1;
			schedule();
                        // We are allowed to continue here because the
                        // VM_FLUSH flag tells us that we are swapping out our
                        // *own* address space, so it must still be valid.
		}
	}
	return 0;
}

/*
 * Called to push all the pages of a process out of memory, typically
 * before a call to mlockall().
 */
void swap_out_myvm(void) {
    struct vm_area_struct *vma;
    struct mm_struct *mm = current->mm;
    mm->def_flags |= VM_FLUSH;
    for (vma = mm->mmap; vma; vma = vma->vm_next) {
        vma->vm_flags |= VM_FLUSH;
        swap_out_vma(current, vma, pgd_offset(mm, vma->vm_start), vma->vm_start, 0, 1);
        vma->vm_flags &= ~VM_FLUSH;
    }
    mm->def_flags &= ~VM_FLUSH;
}

static int swap_out_process(struct task_struct * p, int dma, int wait)
{
	unsigned long address;
	struct vm_area_struct* vma;

        /*
         * If a process is being shredded, ignore it.
         */
        if (p->mm->def_flags & VM_FLUSH)
            return 0;

	/*
	 * Go through process' page directory.
	 */
	address = p->swap_address;
	p->swap_address = 0;

	/*
	 * Find the proper vm-area
	 */
	vma = find_vma(p->mm, address);
	if (!vma)
		return 0;
	if (address < vma->vm_start)
		address = vma->vm_start;

	for (;;) {
		int result = swap_out_vma(p, vma, pgd_offset(p->mm, address), address, dma, wait);
		if (!(vma->vm_flags & VM_FLUSH) && result)
			return result;
		vma = vma->vm_next;
		if (!vma)
			break;
		address = vma->vm_start;
		if (need_resched) {
			if (!(vma->vm_flags & VM_FLUSH))
				return -1;
			schedule();
                        // We are allowed to continue here because the
                        // VM_FLUSH flag tells us that we are swapping out our
                        // *own* address space, so it must still be valid.
		}
	}
	p->swap_address = 0;
	return 0;
}

#if 0
struct task_struct *swaplist[NR_TASKS];
int ntask;

#define swapone(i,j)  {\
	struct task_struct *p = swaplist[i]; \
	swaplist[i] = swaplist[j]; \
	swaplist[j] = p; }

void
make_swap_list(void) {
	int i,j;

	/* generate list of eligible processes */
	ntask = 0;
	for (i = 1; i < NR_TASKS; i++) {
		struct task_struct *p = task[i];
		if (p) {
			if (p->swappable && p->mm->rss) {
				swaplist[ntask++] = p;
			}
		}
	}

	/* classic bubble sort; inverse priority order */
	for (i = 0; i < ntask - 1; i++) {
		for (j = i + 1; j < ntask; j++) {
			if (swaplist[i]->policy != SCHED_OTHER) {

				/* real-time processes get priority ... */
				if (swaplist[j]->policy != SCHED_OTHER) {

					/* unless the other is a real-time process .. */
					if (swaplist[i]->rt_priority > swaplist[j]->rt_priority) {

						/* but it wins, it has a higher priority */
						swapone(i, j);
					}

					/* same priority, different policies ... */
					if (swaplist[i]->rt_priority == swaplist[j]->rt_priority) {

						/* run-to-completion has precedence ... */
						if (swaplist[i]->policy == SCHED_FIFO) {

							/* so he wins */
							if (swaplist[j]->policy == SCHED_RR) {
								swapone(i, j);
							}
						}
					}
				}

				/* non-real-time other always loses */
				else {
					swapone(i, j);
				}
			}
			else {
				/* real time other always wins; if not ... */
				if (swaplist[j]->policy == SCHED_OTHER) {

					/* test priority, whatever that means in Linux */
					if (swaplist[j]->priority < swaplist[i]->priority) {
						swapone(i, j);
					}
				}
			}
		}
	}
}

#ifdef CONFIG_XMON
static int
countpages(struct mm_struct *mm) {
    struct vm_area_struct *vma;
    int cnt = 0;
    for (vma = mm->mmap; vma; vma = vma->vm_next) {
        extern pte_t *findPTE(struct mm_struct *, unsigned long);
        unsigned long saddr = vma->vm_start;
        for (; saddr < vma->vm_end; saddr += PAGE_SIZE)
            if (findPTE(mm, saddr))
                cnt++;
    }
    return cnt;
}

void
dump_swaplist(void) {
	int i;
	for (i = 0; i < ntask; i++) {
		struct task_struct *p = swaplist[i];
        int cnt = countpages(p->mm);
		printk("%p %5d: %s %3ld %3ld pages %d rss %ld\n",
			p,
			p->pid,
			(p->policy == SCHED_FIFO ? "fifo" :
			 (p->policy == SCHED_RR ? "rr" : "other")),
			p->rt_priority,
			p->priority,
            cnt,
            p->mm->rss);
	}
}
#endif

static int swap_out(unsigned int priority, int dma, int wait)
{
	struct task_struct **p, **pe;

	make_swap_list();
	p = &swaplist[0];
	pe = &swaplist[ntask];
	while (p < pe) {
		if (!(*p)->swap_cnt) {
 			/* Normalise the number of pages swapped by
			   multiplying by (RSS / 1MB) */
			(*p)->swap_cnt = AGE_CLUSTER_SIZE((*p)->mm->rss);
		}
		else if (--(*p)->swap_cnt) {
			if (swap_out_process((*p), dma, wait)) {
				return 1;
			}
		}
		p++;
	}
	return 0;
}

int _try_to_free_page(int priority, int dma) {
		if (shrink_mmap(priority, dma)) {
			return 1;
		}
		schedule();
		if (kmem_cache_reap(priority, dma, 0)) {
			return 1;
		}
		schedule();
		if (shm_swap(priority, dma)) {
			return 1;
		}
		schedule();
		if (swap_out(priority, dma, 0)) {
			return 1;
		}
        return 0;
}
#else
void
dump_swaplist(void) {}

static int swap_out(unsigned int priority, int dma, int wait)
{
	static int swap_task;
	int loop, counter;
	struct task_struct *p;
        int rval;

	counter = ((PAGEOUT_WEIGHT * nr_tasks) >> 10) >> priority;
	for(; counter >= 0; counter--) {
		/*
		 * Check that swap_task is suitable for swapping.  If not, look for
		 * the next suitable process.
		 */
		loop = 0;
		while(1) {
			if (swap_task >= NR_TASKS) {
				swap_task = 1;
				if (loop)
					/* all processes are unswappable or already swapped out */
					return 0;
				loop = 1;
			}

			p = task[swap_task];
			if (p &&
			    !(p->flags & PF_EXITING) &&
			    p->swappable &&
			    p->mm->rss &&
			    !(p->mm->def_flags & VM_FLUSH))
				break;

			swap_task++;
		}

		/*
		 * Determine the number of pages to swap from this process.
		 */
		if (!p->swap_cnt) {
 			/* Normalise the number of pages swapped by
			   multiplying by (RSS / 1MB) */
			p->swap_cnt = AGE_CLUSTER_SIZE(p->mm->rss);
		}
		if (!--p->swap_cnt)
			swap_task++;
                rval = swap_out_process(p, dma, wait);
		switch (rval) {
			case 0:
				if (p->swap_cnt)
					swap_task++;
				break;
			case 1:
				return 1;
			case -1:
				schedule();
				break;
			default:
				break;
		}
	}
	return 0;
}
#endif

/*
 * We are much more aggressive about trying to swap out than we used
 * to be.  This works out OK, because we now do proper aging on page
 * contents. 
 */
int try_to_free_page(int priority, int dma, int wait)
{
	static int state = 0;
	int i=3;
	int stop;
	int gotone = 1;

	/* we don't try as hard if we're not waiting.. */
	stop = 3;
	if (wait)
		stop = 0;
	do {
		if (need_resched)
		    schedule();
		switch (state) {
		case 0:
			if (shrink_mmap(i, dma))
				goto EXIT;
			state = 1;
		case 1:
			if (kmem_cache_reap(i, dma, wait))
				goto EXIT;
			state = 2;
		case 2:
			if (shm_swap(i, dma))
				goto EXIT;
			state = 3;
		default:
			if (swap_out(i, dma, wait))
				goto EXIT;
			state = 0;
		}
		i--;
	} while ((i - stop) >= 0);
	gotone = 0;
EXIT:

	return gotone;
}

/*
 * The background pageout daemon.
 * Started as a kernel thread from the init process.
 */
int kswapd(void *unused)
{
	int i;
	char *revision="$Revision: 1.12 $", *s, *e;
	
	current->session = 1;
	current->pgrp = 1;
	sprintf(current->comm, "kswapd");
	current->blocked = ~0UL;
	
	/*
	 *	As a kernel thread we want to tamper with system buffers
	 *	and other internals and thus be subject to the SMP locking
	 *	rules. (On a uniprocessor box this does nothing).
	 */
	lock_kernel();

	/* Give kswapd a realtime priority. */
	current->policy = SCHED_RR;
	current->rt_priority = 1;
	current->priority = DEF_PRIORITY;   /* Fixme --- we need to standardise our
				    namings for POSIX.4 realtime scheduling
				    priorities.  */

	init_swap_timer();
	
	if ((s = strchr(revision, ':')) &&
	    (e = strchr(s, '$')))
		s++, i = e - s;
	else
		s = revision, i = -1;
	printk ("Started kswapd v%.*s\n", i, s);

	while (1) {
		kswapd_awake = 0;
		current->signal = 0;
		run_task_queue(&tq_disk);
		interruptible_sleep_on(&kswapd_wait);
		kswapd_awake = 1;
		swapstats.wakeups++;
		/* Do the background pageout: */
                for (i = 0; i < kswapd_ctl.maxpages; i++) {
                    try_to_free_page(GFP_KERNEL, 0, 0);
                    if (need_resched)
                        break;
                }
	}
}

/* 
 * The swap_tick function gets called on every clock tick.
 */

void swap_tick(void)
{
	int	want_wakeup = 0;
	static int	last_wakeup_low = 0;

	if ((nr_free_pages + nr_async_pages) < free_pages_low) {
		if (last_wakeup_low)
			want_wakeup = jiffies >= next_swap_jiffies;
		else
			last_wakeup_low = want_wakeup = 1;
	}
	else if (((nr_free_pages + nr_async_pages) < free_pages_high) && 
	         jiffies >= next_swap_jiffies) {
		last_wakeup_low = 0;
		want_wakeup = 1;
	}

	if (want_wakeup) {
		if (!kswapd_awake && kswapd_ctl.maxpages > 0) {
			wake_up(&kswapd_wait);
			need_resched = 1;
		}
		next_swap_jiffies = jiffies + swapout_interval;
	}
	timer_active |= (1<<SWAP_TIMER);
}


/* 
 * Initialise the swap timer
 */

void init_swap_timer(void)
{
	timer_table[SWAP_TIMER].expires = 0;
	timer_table[SWAP_TIMER].fn = swap_tick;
	timer_active |= (1<<SWAP_TIMER);
}
